# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import math
from collections import namedtuple
import blockdiag.metrics
from nwdiag import elements
from blockdiag.utils import Box, XY

cellsize = blockdiag.metrics.DiagramMetrics.cellsize


class DiagramMetrics(blockdiag.metrics.DiagramMetrics):
    node_width = cellsize * 13
    span_width = cellsize * 6
    span_height = cellsize * 13

    def __init__(self, diagram, **kwargs):
        super(DiagramMetrics, self).__init__(diagram, **kwargs)

        self.networks = diagram.networks
        self.trunk_diameter = self.cellsize
        self.jump_shift = self.trunk_diameter // 2
        self.jump_radius = self.trunk_diameter
        self.page_padding = [self.span_height // 2, 0, 0, self.node_width]

        for node in diagram.nodes:
            bottom = [n for n in node.networks if n.xy.y > node.xy.y]
            cnwidth = (len(bottom) + 1) * self.cellsize * 2
            if self.cell(node).width < cnwidth:
                node.width = cnwidth
                self.spreadsheet.set_node_width(node.xy.x, cnwidth)

    def node(self, node):
        n = super(DiagramMetrics, self).cell(node)
        return NodeMetrics(node, self, n.x1, n.y1, n.x2, n.y2)

    def cell(self, node):
        if isinstance(node, elements.Network):
            metrics = super(DiagramMetrics, self).cell(node, use_padding=False)
        elif isinstance(node, elements.NodeGroup):
            n = super(DiagramMetrics, self).cell(node)
            metrics = GroupMetrics(node, self, n.x1, n.y1, n.x2, n.y2)
        else:
            metrics = super(DiagramMetrics, self).cell(node)
        return metrics

    def network(self, network):
        n = self.cell(network)
        return NetworkMetrics(self, n.x1, n.y1, n.x2, n.y2)


class NetworkMetrics(blockdiag.metrics.NodeMetrics):
    @property
    def trunkline(self):
        x0 = self.left.x
        x1 = self.right.x
        y = self.top.y

        return [XY(x0, y), XY(x1, y)]

    @property
    def top(self):
        pt = self.box.top
        return XY(pt.x, pt.y - self.span_height // 2)

    @property
    def left(self):
        pt = self.box.left
        return XY(pt.x - self.span_width // 2, pt.y)

    @property
    def right(self):
        pt = self.box.right
        return XY(pt.x + self.span_width // 2, pt.y)

    @property
    def textbox(self):
        x = self.left.x
        y = self.top.y

        width = self.node_width * 3 // 2
        height = self.node_height

        return Box(x - width, y - height // 2, x, y + height // 2)


class NodeMetrics(blockdiag.metrics.NodeMetrics):
    def __init__(self, node, metrics, x1, y1, x2, y2):
        super(NodeMetrics, self).__init__(metrics, x1, y1, x2, y2)
        self.node = node

    @property
    def connectors(self):
        above = [n for n in self.node.networks if n.xy.y <= self.node.xy.y]
        above.sort(key=lambda a: a.xy.y, reverse=True)

        bottom = [n for n in self.node.networks if n.xy.y > self.node.xy.y]
        bottom.sort(key=lambda a: a.xy.y)

        Connector = namedtuple('Connector', 'network line textbox')

        connectors = []
        for networks in [above, bottom]:
            for network in networks:
                if network.hidden:
                    span = 0
                else:
                    span = self.trunk_diameter // 2

                if network.xy.y <= self.node.xy.y:
                    x, y2 = self.top
                    y1 = self.network(network).top.y + span
                else:
                    x, y1 = self.bottom
                    y2 = self.network(network).top.y - span

                if len(networks) == 1:
                    dx = 0
                else:
                    pos = networks.index(network)
                    base_x = (len(networks) - 1) / 2.0 - pos
                    dx = int(math.floor(base_x * self.cellsize * 2))

                width = self.node_width + self.span_width
                textbox = Box(x + dx + self.cellsize // 2,
                              y2 - self.span_height // 2,
                              x + width - self.cellsize // 2,
                              y2)
                line = [XY(x + dx, y1), XY(x + dx, y2)]

                cn = Connector(network, line, textbox)
                connectors.append(cn)

        return connectors


class GroupMetrics(blockdiag.metrics.NodeMetrics):
    def __init__(self, group, metrics, x1, y1, x2, y2):
        super(GroupMetrics, self).__init__(metrics, x1, y1, x2, y2)

        self.is_root_group = False
        if group.nodes:
            networks = group.nodes[0].networks[:]
            networks.sort(key=lambda a: a.xy.y)
            network = min(networks)
            if self.top.x == metrics.network(network).top.x:
                self.is_root_group = True

    @property
    def grouplabelbox(self):
        box = super(GroupMetrics, self).grouplabelbox
        span = self.cellsize
        box = Box(box[0], box[1] + span, box[2], box[3] + span)

        if self.is_root_group:
            width = (self.node_width + self.span_width) // 2
            box = Box(box[0] + width, box[1], box[2] + width, box[3])

        return box

    @property
    def marginbox(self):
        box = super(GroupMetrics, self).box
        margin_x = self.span_height // 2 - self.cellsize
        margin_y = self.cellsize
        return Box(box[0] - margin_y, box[1] - margin_x,
                   box[2] + margin_y, box[3] + margin_x)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from rackdiag import parser
from rackdiag.elements import Diagram, Rack, RackItem
from blockdiag.utils import unquote, XY


class DiagramTreeBuilder:
    def build(self, tree):
        self.diagram = Diagram()
        self.instantiate(self.diagram.racks[0], tree)

        for rack in self.diagram.racks[:]:
            if len(rack.nodes) == 0:
                self.diagram.racks.remove(rack)

        return self.diagram

    def instantiate(self, rack, tree):
        for stmt in tree.stmts:
            if isinstance(stmt, parser.Attr):
                try:
                    rack.set_attribute(stmt)
                except AttributeError:
                    self.diagram.set_attribute(stmt)
            elif isinstance(stmt, parser.RackItem):
                if stmt.number is None:
                    if len(rack.nodes) > 0:
                        item = rack.nodes[-1]
                        number = item.number + item.colheight
                    else:
                        number = 1
                else:
                    number = stmt.number

                item = RackItem(number, unquote(stmt.label))
                item.set_attributes(stmt.attrs)
                rack.nodes.append(item)
            elif isinstance(stmt, parser.Rack):
                _rack = Rack()
                self.diagram.racks.append(_rack)
                self.instantiate(_rack, stmt)
            elif isinstance(stmt, parser.AttrPlugin):
                self.diagram.set_plugin(stmt.name, stmt.attrs)


class DiagramLayoutManager:
    def __init__(self, diagram):
        self.diagram = diagram

    def run(self):
        x = 0
        for rack in self.diagram.racks:
            self.layout_rack(rack)
            rack.xy = XY(x, 0)
            rack.fixiate()

            x += rack.colwidth

        self.diagram.fixiate()

    def layout_rack(self, rack):
        usage = defaultdict(bool)

        for item in rack.nodes:
            item.xy = XY(-1, -1)

        for item in rack.nodes:
            if rack.descending:
                y = rack.colheight - item.number - item.colheight + 1
            else:
                y = item.number - 1

            for x in range(255):
                r = range(y, y + item.colheight)
                if len([_ for _ in r if usage[(x, _)]]) == 0:
                    break

            item.xy = XY(x, y)
            for dy in range(y, y + item.colheight):
                usage[(x, dy)] = 1
            self.validate_rack(rack, item)

    def validate_rack(self, rack, item):
        if item.xy.y < 0:
            msg = "Rack %d is layouted to underground!\n" % item.number
            raise AttributeError(msg)
        elif rack.colheight < item.xy.y + item.colheight:
            msg = "Rack %d is oversized to rack-height\n" % item.number
            raise AttributeError(msg)


class ScreenNodeBuilder:
    @classmethod
    def build(cls, tree):
        Rack.clear()
        RackItem.clear()
        Diagram.clear()

        return cls(tree).run()

    def __init__(self, tree):
        self.diagram = DiagramTreeBuilder().build(tree)

    def run(self):
        DiagramLayoutManager(self.diagram).run()
        return self.diagram



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import re
import blockdiag.elements
from rackdiag.utils.math import lcm
from blockdiag.utils import XY
from blockdiag.utils.compat import u


class DiagramNode(blockdiag.elements.DiagramNode):
    pass


class RackItem(blockdiag.elements.DiagramNode):
    desctable = []
    attrname = {}

    @classmethod
    def clear(cls):
        super(RackItem, cls).clear()
        cls.desctable = ['number', 'label', 'units', 'ampere',
                         'weight', 'description']
        cls.attrname = dict(number='No', label='Name', units='Height',
                            ampere='Capacity', weight='Weight',
                            description='Description')

    def __init__(self, number, label):
        super(RackItem, self).__init__(None)
        self.number = int(number)
        self.label = label
        self.ampere = None
        self.weight = None

        if self.label == 'N/A':
            self.color = 'gray'

    @property
    def display_label(self):
        attrs = []
        if self.colheight > 1:
            attrs.append(u("%dU") % self.colheight)
        if self.ampere:
            attrs.append(u("%.1fA") % self.ampere)
        if self.weight:
            attrs.append(u("%.1fkg") % self.weight)

        labels = []
        if self.label:
            labels.append(self.label)
        if attrs:
            labels.append(u("[%s]") % u("/").join(attrs))

        return u("\n").join(labels)

    def set_attribute(self, attr):
        if re.search('^\d+U$', attr.name):
            self.colheight = int(attr.name[:-1])
        elif re.search('^\d+(\.[0-9]+)?A$', attr.name):
            self.ampere = float(attr.name[:-1])
        elif re.search('^\d+(\.[0-9]+)?kg$', attr.name):
            self.weight = float(attr.name[:-2])
        else:
            return super(RackItem, self).set_attribute(attr)

    def set_height(self, value):
        self.colheight = int(value)

    def to_desctable(self):
        attrs = []
        for name in self.desctable:
            if name == 'units':
                attrs.append(u("%dU") % self.colheight)
            elif name == 'ampere':
                if self.ampere is None:
                    attrs.append(u(""))
                else:
                    attrs.append((u("%.1fA") % self.ampere) or u(""))
            elif name == 'weight':
                if self.weight is None:
                    attrs.append(u(""))
                else:
                    attrs.append((u("%.1fkg") % self.weight) or u(""))
            else:
                value = getattr(self, name)
                if value is None:
                    attrs.append(u(""))
                elif isinstance(value, int):
                    attrs.append(str(value))
                else:
                    attrs.append(getattr(self, name))

        return attrs


class Rack(blockdiag.elements.NodeGroup):
    def __init__(self):
        super(Rack, self).__init__(None)
        self.colheight = 42
        self.description = None
        self.descending = True

    def set_default_fontsize(self, value):
        raise AttributeError()  # Ignore and pass-through to Diagram

    @property
    def display_label(self):
        attrs = []
        if self.description or attrs:
            attrs.insert(0, u("%dU") % self.colheight)

        labels = []
        if self.description:
            labels.append(self.description)
        if attrs:
            labels.append(u("[%s]") % u("/").join(attrs))

        return u("\n").join(labels)

    def items(self, levels):
        if isinstance(levels, int):
            levels = [levels]

        return [node for node in self.nodes if node.xy.y in levels]

    def get_linked_levels(self, level):
        def get_max_height(self, y):
            try:
                return max(n.colheight for n in self.items(y))
            except:
                return 0

        height = get_max_height(self, level)
        for dy in range(1, self.colheight - level + 1):
            if dy < height:
                height = max(height, get_max_height(self, level + dy) + dy)

        return range(level, level + height)

    def adjust_node_widths(self):
        i = 0
        linked_widths = {}
        widths = []
        while i < self.colheight:
            levels = self.get_linked_levels(i)

            if levels:
                linked_widths[i] = []
                for level in levels:
                    nodes = self.items(level)
                    if nodes:
                        width = max(n.xy.x for n in nodes) + 1
                        widths.append(width)
                        linked_widths[i].append(width)

                i += len(levels)
            else:
                i += 1

        i = 0
        self.colwidth = lcm(*widths) or 1
        while i < self.colheight:
            levels = self.get_linked_levels(i)

            if levels:
                colwidth = max(linked_widths[i]) or 1
                for level in levels:
                    nodes = self.items(level)
                    if nodes:
                        width = self.colwidth // colwidth
                        for node in nodes:
                            node.xy = XY(node.xy.x * width, node.xy.y)
                            node.colwidth = width

                i += len(levels)
            else:
                i += 1

    def fixiate(self):
        self.adjust_node_widths()

        for node in self.nodes:
            node.xy = node.xy.shift(x=self.xy.x)

    def set_ascending(self, attr):
        self.descending = False

    def set_attribute(self, attr):
        if re.search('^\d+U$', attr.name):
            self.colheight = int(attr.name[:-1])
        else:
            return super(Rack, self).set_attribute(attr)


class Diagram(blockdiag.elements.Diagram):
    _DiagramNode = RackItem

    @classmethod
    def clear(cls):
        super(Diagram, cls).clear()
        cls._DiagramNode.clear()

    def __init__(self):
        super(Diagram, self).__init__()

        self.racks = [Rack()]

    def set_default_fontsize(self, value):
        super(Diagram, self).set_default_fontsize(value)
        self.fontsize = int(value)

    def set_rackheight(self, value):
        self.racks[0].colheight = int(value)

    def fixiate(self):
        self.colwidth = sum(r.colwidth for r in self.racks)
        self.colheight = max(r.colheight for r in self.racks)

    def traverse_nodes(self, **kwargs):
        for rack in self.racks:
            for node in rack.traverse_nodes():
                yield node



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import blockdiag.metrics
from collections import defaultdict, namedtuple
from blockdiag.utils import Box, XY


class DiagramMetrics(blockdiag.metrics.DiagramMetrics):
    def __init__(self, diagram, **kwargs):
        span_height = self.span_height
        span_width = self.span_width
        self.span_height = 0
        self.span_width = 0
        self.node_width = self.cellsize * 3
        self.colwidth = diagram.colwidth
        super(DiagramMetrics, self).__init__(diagram, **kwargs)

        # reset node_width FORCE
        self.spreadsheet.node_width = defaultdict(lambda: self.node_width)

        self.spreadsheet.set_span_height(0, span_height * 2)
        self.spreadsheet.set_span_height(diagram.colheight, span_height)
        self.spreadsheet.set_span_width(0, span_width)
        self.spreadsheet.set_span_width(diagram.colwidth, span_width)

    def measure_line(self, n):
        _Node = namedtuple('Node', 'xy')

        if n == self.colwidth:
            node = _Node(XY(n - 1, 0))
            pt = self.spreadsheet._node_topleft(node, use_padding=False)
            pt = pt.shift(x=self.node_width)
        else:
            node = _Node(XY(n, 0))
            pt = self.spreadsheet._node_topleft(node, use_padding=False)

        if n * 2.0 % self.colwidth == 0:
            return (XY(pt.x, pt.y - self.cellsize * 4), pt)
        elif n * 4.0 % self.colwidth == 0:
            return (XY(pt.x, pt.y - self.cellsize * 4), pt)
        else:
            return (XY(pt.x, pt.y - self.cellsize * 2), pt)

    def measure_label(self, n):
        top, _ = self.measure_line(n)
        cellsize = self.cellsize

        return Box(top.x - cellsize * 4, top.y - cellsize * 3,
                   top.x + cellsize * 4, top.y)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from docutils.parsers import rst
import nwdiag.parser
import nwdiag.builder
import nwdiag.drawer
from nwdiag.utils.rst.nodes import nwdiag as nwdiag_node
from blockdiag.utils.rst import directives


directive_options_default = dict(format='PNG',
                                 antialias=False,
                                 fontpath=None,
                                 outputdir=None,
                                 nodoctype=False,
                                 noviewbox=False,
                                 inline_svg=False,
                                 ignore_pil=False)
directive_options = {}


class NwdiagDirective(directives.BlockdiagDirective):
    name = "nwdiag"
    node_class = nwdiag_node
    processor = nwdiag

    @property
    def global_options(self):
        return directive_options


def setup(**kwargs):
    for key, value in directive_options_default.items():
        directive_options[key] = kwargs.get(key, value)

    rst.directives.register_directive("nwdiag", NwdiagDirective)



# -*- coding: utf-8 -*-

# Copyright (c) 2008/2009 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

r'''A DOT language parser using funcparserlib.

The parser is based on [the DOT grammar][1]. It is pretty complete with a few
not supported things:

* Ports and compass points
* XML identifiers

At the moment, the parser builds only a parse tree, not an abstract syntax tree
(AST) or an API for dealing with DOT.

  [1]: http://www.graphviz.org/doc/info/lang.html
'''

import io
from re import MULTILINE, DOTALL
from collections import namedtuple
from funcparserlib.lexer import make_tokenizer, Token, LexerError
from funcparserlib.parser import (some, a, maybe, many, finished, skip)
from blockdiag.parser import create_mapper, oneplus_to_list
from blockdiag.utils.compat import u


Diagram = namedtuple('Diagram', 'id stmts')
RackItem = namedtuple('RackItem', 'number label attrs')
Attr = namedtuple('Attr', 'name value')
Rack = namedtuple('Rack', 'id stmts')
Extension = namedtuple('Extension', 'type name attrs')


class ParseException(Exception):
    pass


def tokenize(string):
    """str -> Sequence(Token)"""
    # flake8: NOQA
    specs = [                                                                                # NOQA
        ('Comment',        (r'/\*(.|[\r\n])*?\*/', MULTILINE)),                              # NOQA
        ('Comment',        (r'(//|#).*',)),                                                  # NOQA
        ('NL',             (r'[\r\n]+',)),                                                   # NOQA
        ('QuotedRackItem', (r'(?<=[:*\-])\s*(?P<quote>"|\').*?(?<!\\)(?P=quote)', DOTALL)),  # NOQA
        ('RackItem',       (r'(?<=[:*\-])[^\r\n\[;}]+',)),                                   # NOQA
        ('Space',          (r'[ \t\r\n]+',)),                                                # NOQA
        ('RackHeight',     (r'[0-9]+U',)),                                                   # NOQA
        ('Units',          (r'[0-9]+(?:\.[0-9]+)?(A|kg)',)),                                 # NOQA
        ('Number',         (r'[0-9]+',)),                                                    # NOQA
        ('Name',           (u('[A-Za-z_0-9\u0080-\uffff]') +                                 # NOQA
                            u('[A-Za-z_\\-.0-9\u0080-\uffff]*'),)),                          # NOQA
        ('Op',             (r'[{}:;,*\-=\[\]]',)),                                           # NOQA
        ('String',         (r'(?P<quote>"|\').*?(?<!\\)(?P=quote)', DOTALL)),                # NOQA
    ]
    useless = ['Comment', 'NL', 'Space']
    t = make_tokenizer(specs)
    return [x for x in t(string) if x.type not in useless]


def parse(seq):
    """Sequence(Token) -> object"""
    id_tokens = ['Name', 'Number', 'String', 'RackHeight', 'Units']
    rackitem_tokens = ['QuotedRackItem', 'RackItem']

    tokval = lambda x: x.value
    op = lambda s: a(Token('Op', s)) >> tokval
    op_ = lambda s: skip(op(s))
    _id = some(lambda t: t.type in id_tokens) >> tokval
    keyword = lambda s: a(Token('Name', s)) >> tokval
    rackheight = some(lambda t: t.type == 'RackHeight') >> tokval
    number = some(lambda t: t.type == 'Number').named('number') >> tokval
    rackitem = some(lambda t: t.type in rackitem_tokens) >> tokval

    def make_num_rackitem(num, text, attr):
        return RackItem(num, text.strip(), attr)

    def make_nonnum_rackitem(text, attr):
        return RackItem(None, text.strip(), attr)

    def make_nonvalued_attr(rackheight):
        return Attr(rackheight, None)

    #
    # parts of syntax
    #
    option_stmt = (
        _id +
        maybe(op_('=') + _id)
        >> create_mapper(Attr)
    )
    option_list = (
        maybe(op_('[') + option_stmt + many(op_(',') + option_stmt) + op_(']'))
        >> create_mapper(oneplus_to_list, default_value=[])
    )

    #  attributes statement::
    #     default_shape = box;
    #     default_fontsize = 16;
    #     12U;
    #     ascending;
    #
    attribute_stmt = (
        _id + op_('=') + _id
        >> create_mapper(Attr)
    )
    rackheight_stmt = (
        rackheight
        >> make_nonvalued_attr
    )
    ascending_stmt = (
        keyword('ascending')
        >> make_nonvalued_attr
    )

    #  field statement::
    #     1: A
    #     2: B [attr = value, attr = value];
    #     * C [attr = value, attr = value];
    #     * D [attr = value, attr = value];
    #
    numbered_field_item_stmt = (
        number +
        op_(':') +
        rackitem +
        option_list
        >> create_mapper(make_num_rackitem)
    )
    nonnumbered_field_item_stmt = (
        (op_('-') | op_('*')) +
        rackitem +
        option_list
        >> create_mapper(make_nonnum_rackitem)
    )
    field_item_stmt = (
        numbered_field_item_stmt |
        nonnumbered_field_item_stmt
    )

    #  rack statement::
    #     rack {
    #       1: A;
    #     }
    #

    # rack definition
    rack_inline_stmt = (
        attribute_stmt |
        rackheight_stmt |
        field_item_stmt
    )
    rack_inline_stmt_list = (
        many(rack_inline_stmt + skip(maybe(op(';'))))
    )
    rack_stmt = (
        skip(keyword('rack')) +
        maybe(_id) +
        op_('{') +
        rack_inline_stmt_list +
        op_('}')
        >> create_mapper(Rack)
    )

    #  extension statement (plugin)::
    #     plugin attributes [name = Name];
    #
    extension_stmt = (
        keyword('plugin') +
        _id +
        option_list
        >> create_mapper(Extension)
    )

    #
    # diagram statement::
    #     rackdiag {
    #        A;
    #     }
    #
    diagram_id = (
        (keyword('diagram') | keyword('rackdiag')) +
        maybe(_id)
        >> list
    )
    diagram_inline_stmt = (
        ascending_stmt |
        extension_stmt |
        rack_stmt |
        field_item_stmt |
        attribute_stmt |
        rackheight_stmt
    )
    diagram_inline_stmt_list = (
        many(diagram_inline_stmt + skip(maybe(op(';'))))
    )
    diagram = (
        maybe(diagram_id) +
        op_('{') +
        diagram_inline_stmt_list +
        op_('}')
        >> create_mapper(Diagram)
    )
    dotfile = diagram + skip(finished)

    return dotfile.parse(seq)


def sort_tree(tree):
    def weight(node):
        if isinstance(node, (Attr, Extension)):
            return 1
        else:
            return 2

    if hasattr(tree, 'stmts'):
        tree.stmts.sort(key=lambda x: weight(x))
        for stmt in tree.stmts:
            sort_tree(stmt)

    return tree


def parse_string(string):
    try:
        tree = parse(tokenize(string))
        return sort_tree(tree)
    except LexerError as e:
        message = "Got unexpected token at line %d column %d" % e.place
        raise ParseException(message)
    except Exception as e:
        raise ParseException(str(e))


def parse_file(path):
    code = io.open(path, 'r', encoding='utf-8-sig').read()
    return parse_string(code)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = [
    'core', 'utils'
]

import packetdiag.parser
import packetdiag.builder
import packetdiag.drawer
core = packetdiag

import packetdiag.utils.rst.nodes
import packetdiag.utils.rst.directives
utils = packetdiag.utils

import blockdiag.utils.bootstrap
utils.bootstrap = blockdiag.utils.bootstrap

import blockdiag.utils.compat
utils.compat = blockdiag.utils.compat

import blockdiag.utils.fontmap
utils.fontmap = blockdiag.utils.fontmap



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import blockdiag.drawer
from packetdiag.metrics import DiagramMetrics
from blockdiag.utils.compat import u


class DiagramDraw(blockdiag.drawer.DiagramDraw):
    def create_metrics(self, *args, **kwargs):
        return DiagramMetrics(*args, **kwargs)

    def _draw_background(self):
        # do not call blockdiag.DiagramDraw#_draw_background()

        scale_interval = self.diagram.scale_interval
        if scale_interval is None:
            scale_interval = self.diagram.colwidth / 2

        # draw measure lines and labels
        font = self.metrics.font_for(None)
        for i in range(self.diagram.colwidth + 1):
            line = self.metrics.measure_line(i)
            self.drawer.line(line, fill=self.diagram.linecolor)

            if (i % scale_interval) == 0:
                box = self.metrics.measure_label(i)

                if self.diagram.scale_direction == "left_to_right":
                    label = u(str(i))
                else:
                    label = u(str(self.diagram.colwidth - i))

                self.drawer.textarea(box, label, font,
                                     fill=self.diagram.textcolor)



# -*- coding: utf-8 -*-
#  Copyright 2012 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from docutils.parsers import rst
from packetdiag.elements import FieldItem
from packetdiag.utils.rst.nodes import packetdiag as packetdiag_node
from blockdiag.utils.rst import directives
from blockdiag.utils.compat import cmp_to_key


directive_options_default = dict(format='PNG',
                                 antialias=False,
                                 fontpath=None,
                                 outputdir=None,
                                 nodoctype=False,
                                 noviewbox=False,
                                 inline_svg=False,
                                 ignore_pil=False)
directive_options = {}


class PacketdiagDirective(directives.BlockdiagDirective):
    name = "packetdiag"
    node_class = packetdiag_node

    @property
    def global_options(self):
        return directive_options

    def description_table(self, diagram):
        nodes = diagram.traverse_nodes
        widths = [25] + [50] * (len(FieldItem.desctable) - 1)
        headers = [FieldItem.attrname[name] for name in FieldItem.desctable]

        descriptions = [n.to_desctable() for n in nodes()]
        descriptions.sort(key=cmp_to_key(directives.cmp_node_number))

        for i in range(len(headers) - 1, -1, -1):
            if any(desc[i] for desc in descriptions):
                pass
            else:
                widths.pop(i)
                headers.pop(i)
                for desc in descriptions:
                    desc.pop(i)

        return self._description_table(descriptions, widths, headers)


def setup(**kwargs):
    for key, value in directive_options_default.items():
        directive_options[key] = kwargs.get(key, value)

    rst.directives.register_directive("packetdiag", PacketdiagDirective)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from packetdiag import parser
from packetdiag.elements import Diagram, FieldItem, DiagramNode
from blockdiag.utils import unquote, XY


class DiagramTreeBuilder:
    def build(self, tree):
        self.diagram = Diagram()
        self.instantiate(tree)

        return self.diagram

    def instantiate(self, tree):
        for stmt in tree.stmts:
            if isinstance(stmt, parser.Attr):
                self.diagram.set_attribute(stmt)
            elif isinstance(stmt, parser.FieldItem):
                item = FieldItem(stmt.begin, stmt.end, unquote(stmt.label))
                item.set_attributes(stmt.attrs)

                if item.number is None:
                    if len(self.diagram.fields) == 0:
                        item.number = 0
                    else:
                        last_item = self.diagram.fields[-1]
                        item.number = last_item.number + last_item.colwidth

                self.diagram.fields.append(item)

            elif isinstance(stmt, parser.AttrPlugin):
                self.diagram.set_plugin(stmt.name, stmt.attrs)


class DiagramLayoutManager:
    def __init__(self, diagram):
        self.diagram = diagram

    def split_field_by_column(self):
        for field in self.diagram.fields:
            while True:
                x = field.number % self.diagram.colwidth
                if x + field.colwidth <= self.diagram.colwidth:
                    break
                else:
                    colwidth = self.diagram.colwidth - x

                    splitted = field.duplicate()
                    splitted.colwidth = colwidth
                    if self.diagram.scale_direction == "left_to_right":
                        splitted.separated_right = True
                    else:
                        splitted.separated_left = True

                    self.diagram.fields.append(splitted)

                    field.number += colwidth
                    field.colwidth -= colwidth
                    if self.diagram.scale_direction == "left_to_right":
                        field.separated_left = True
                    else:
                        field.separated_right = True

            yield field

    def run(self):
        filled = {}
        for field in self.split_field_by_column():
            x = field.number % self.diagram.colwidth
            y = field.number // self.diagram.colwidth

            if filled.get(y) is None:
                filled[y] = {}

            for rx in range(x, x + field.colwidth):
                if filled[y].get(rx):
                    msg = ("Field '%s' is conflicted to other field\n" %
                           field.label)
                    raise AttributeError(msg)
                filled[y][rx] = True

            if self.diagram.scale_direction == "right_to_left":
                x = self.diagram.colwidth - x - field.colwidth

            field.xy = XY(x, y)

        self.diagram.fixiate()


class ScreenNodeBuilder:
    @classmethod
    def build(cls, tree):
        DiagramNode.clear()
        Diagram.clear()

        return cls(tree).run()

    def __init__(self, tree):
        self.diagram = DiagramTreeBuilder().build(tree)

    def run(self):
        DiagramLayoutManager(self.diagram).run()
        return self.diagram



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import blockdiag.elements


class DiagramNode(blockdiag.elements.DiagramNode):
    pass


class FieldItem(blockdiag.elements.DiagramNode):
    shape = '_packet_node'
    desctable = ['number', 'label', 'description']
    attrname = dict(number='Seq', label='Name', description='Description')

    def __init__(self, begin, end, label):
        super(FieldItem, self).__init__(None)
        self.label = label
        self.separated_right = False
        self.separated_left = False

        if begin is None:
            self.number = None
            self.colwidth = 1
        elif end is None:
            self.number = int(begin)
            self.colwidth = 1
        else:
            self.number = int(begin)
            self.colwidth = int(end) - int(begin) + 1

            if self.colwidth <= 0:
                msg = ("Invalid field size definition: %d-%d: %s\n" %
                       (begin, end, label))
                raise AttributeError(msg)

    def set_height(self, value):
        self.colheight = int(value)

    def set_len(self, value):
        self.colwidth = int(value)


class Diagram(blockdiag.elements.Diagram):
    _DiagramNode = FieldItem

    def __init__(self):
        super(Diagram, self).__init__()
        self.colwidth = 16
        self.scale_interval = None
        self.scale_direction = "left_to_right"

        self.int_attrs.append('scale_interval')

    @property
    def fields(self):
        return self.nodes

    def set_scale_direction(self, value):
        value = value.lower()
        if value in ('ltr', 'left_to_right'):
            self.scale_direction = 'left_to_right'
        elif value in ('rtl', 'right_to_left'):
            self.scale_direction = 'right_to_left'
        else:
            msg = "unknown scale_direction: %s\n" % value
            raise AttributeError(msg)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division


def gcd(*args):
    if len(args) == 0:
        return 0
    elif len(args) == 1:
        return args[0]
    elif len(args) == 2:
        if args[1] == 0:
            return args[0]
        else:
            return gcd(args[1], args[0] % args[1])
    else:
        return gcd(gcd(*args[:-1]), args[-1])


def lcm(*args):
    if len(args) == 0:
        return 0
    elif len(args) == 1:
        return args[0]
    elif len(args) == 2:
        return args[0] * args[1] // gcd(*args)
    else:
        return lcm(lcm(*args[:-1]), args[-1])



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from collections import defaultdict
import blockdiag.metrics
from blockdiag.utils import Box, XY
from rackdiag import elements


class DiagramMetrics(blockdiag.metrics.DiagramMetrics):
    def __init__(self, diagram, **kwargs):
        span_height = self.span_height
        span_width = self.span_width
        self.span_height = 0
        self.span_width = 0
        super(DiagramMetrics, self).__init__(diagram, **kwargs)

        labelsize = [self.textsize(r.display_label, font=self.font_for(r))
                     for r in diagram.racks if r.display_label]
        if labelsize:
            labelheight = (max(size.height for size in labelsize) +
                           self.line_spacing * 2)
        else:
            labelheight = 0

        # reset node_width FORCE
        self.spreadsheet.node_width = defaultdict(lambda: self.node_width)

        self.spreadsheet.set_span_height(0, span_height + labelheight)
        self.spreadsheet.set_span_height(diagram.colheight, span_height)
        self.spreadsheet.set_span_width(0, span_width)

        for rack in diagram.racks:
            x = rack.xy.x + rack.colwidth
            self.spreadsheet.set_span_width(x, span_width)

            if rack.colwidth > 1:
                node_width = self.node_width // rack.colwidth
                for i in range(rack.colwidth):
                    self.spreadsheet.set_node_width(rack.xy.x + i, node_width)

    def racklabel(self, rack):
        cell = self.cell(rack)
        textsize = self.textsize(rack.display_label, font=self.font_for(rack))
        y1 = cell.y1 - textsize.height - self.line_spacing * 2
        return Box(cell.x1, y1, cell.x2, cell.y1)

    def racknumber(self, rack, number):
        if rack.descending:
            y = rack.colheight - number - 1
        else:
            y = number

        dummy = elements.DiagramNode(None)
        dummy.xy = XY(rack.xy.x, y)
        dummy.colwidth = 1
        dummy.colheight = 1

        box = self.cell(dummy, use_padding=False).box
        return Box(0, box[1], box[0], box[3])



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import blockdiag.drawer
from rackdiag.metrics import DiagramMetrics
from blockdiag.utils import Box
from blockdiag.utils.compat import u


class DiagramDraw(blockdiag.drawer.DiagramDraw):
    def create_metrics(self, *args, **kwargs):
        return DiagramMetrics(*args, **kwargs)

    def _draw_elements(self, **kwargs):
        default_font = self.metrics.font_for(self.diagram)

        for rack in self.diagram.racks:
            frame = self.metrics.cell(rack, use_padding=False).box
            self.drawer.rectangle(frame, fill='white',
                                  outline=self.diagram.linecolor)

            for i in range(rack.colheight):
                box = self.metrics.racknumber(rack, i)
                number = u("%d") % (i + 1)
                self.drawer.textarea(box, number, default_font, halign='right',
                                     fill=self.diagram.textcolor)

            if rack.display_label:
                box = self.metrics.racklabel(rack)
                self.drawer.textarea(box, rack.display_label,
                                     self.metrics.font_for(rack),
                                     fill=rack.textcolor)

        super(DiagramDraw, self)._draw_elements(**kwargs)

    def _draw_background(self):
        # do not draw shadow of nodes on super()
        self.diagram.shadow_style = 'none'
        super(DiagramDraw, self)._draw_background()

        # draw shadow of frame
        dx, dy = self.metrics.shadow_offset
        for rack in self.diagram.racks:
            frame = self.metrics.cell(rack, use_padding=False)
            shadow = Box(frame.x1 + dx, frame.y1 + dy,
                         frame.x2 + dx, frame.y2 + dy)
            self.drawer.rectangle(shadow, fill=self.shadow, filter='blur')

    def node(self, node, **kwargs):
        label, node.label = node.label, node.display_label
        super(DiagramDraw, self).node(node, **kwargs)
        node.label = label



# -*- coding: utf-8 -*-
from __future__ import print_function

from nwdiag.tests.utils import BuilderTestCase


class TestBuilder(BuilderTestCase):
    def test_diagram_attributes(self):
        diagram = self.build('diagram_attributes.diag')

        self.assertEqual(160, diagram.node_width)
        self.assertEqual(160, diagram.node_height)
        self.assertEqual(32, diagram.span_width)
        self.assertEqual(32, diagram.span_height)
        self.assertEqual(16, diagram.nodes[0].fontsize)
        self.assertEqual('serif', diagram.nodes[0].fontfamily)
        self.assertEqual((255, 0, 0), diagram.networks[0].color)
        self.assertEqual(16, diagram.networks[0].fontsize)
        self.assertEqual('serif', diagram.networks[0].fontfamily)

    def test_node_attributes(self):
        diagram = self.build('node_attributes.diag')

        network = diagram.networks[0]
        assert diagram.nodes[0].address[network] == '192.168.0.1'
        assert diagram.nodes[1].address[network] == '192.168.0.2\n192.168.0.3'

    def test_node_address_attribute(self):
        diagram = self.build('node_address_attribute.diag')

        network = diagram.networks[0]
        self.assertEqual('192.168.0.1',
                         diagram.nodes[0].address[network])
        self.assertEqual('2001:0db8:bd05:01d2:288a:1fc0:0001:10ee',
                         diagram.nodes[1].address[network])

    def test_node_including_hyphen_diagram(self):
        diagram = self.build('node_including_hyphen.diag')

        self.assertEqual('web-01', diagram.nodes[0].id)
        self.assertEqual('web-02', diagram.nodes[1].id)

    def test_single_network_diagram(self):
        diagram = self.build('single_network.diag')

        self.assertEqual(1, len(diagram.nodes))
        self.assertEqual(1, len(diagram.networks))
        self.assertNodeXY(diagram, {'A': (0, 0)})

    def test_two_networks_diagram(self):
        diagram = self.build('two_networks.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1)})

    def test_node_belongs_to_multiple_networks_diagram(self):
        diagram = self.build('node_belongs_to_multiple_networks.diag')
        self.assertNodeXY(diagram, {'A': (0, 0)})

    def test_connected_networks_diagram(self):
        diagram = self.build('connected_networks.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 1)})

    def test_connected_networks2_diagram(self):
        diagram = self.build('connected_networks2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (3, 0),
                                    'E': (4, 0), 'Z': (3, 1)})

    def test_connected_networks3_diagram(self):
        diagram = self.build('connected_networks3.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (3, 0),
                                    'E': (4, 0), 'F': (5, 0),
                                    'G': (6, 0), 'H': (1, 1),
                                    'I': (2, 1), 'J': (4, 2),
                                    'K': (5, 2)})

    def test_group_inner_network_diagram(self):
        diagram = self.build('group_inner_network.diag')

        self.assertEqual(2, len(diagram.groups[0].nodes))
        self.assertEqual(2, diagram.groups[0].colwidth)
        self.assertEqual(1, diagram.groups[0].colheight)
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0)})

    def test_group_outer_network_diagram(self):
        diagram = self.build('group_outer_network.diag')

        self.assertEqual(2, len(diagram.groups[0].nodes))
        self.assertEqual(2, diagram.groups[0].colwidth)
        self.assertEqual(1, diagram.groups[0].colheight)
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0)})

    def test_group_across_network_diagram(self):
        diagram = self.build('group_across_network.diag')

        self.assertEqual(3, len(diagram.groups[0].nodes))
        self.assertEqual(3, diagram.groups[0].colwidth)
        self.assertEqual(2, diagram.groups[0].colheight)
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 1)})

    def test_group_network_diagram(self):
        diagram = self.build('group_network.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (0, 1), 'D': (2, 1),
                                    'E': (3, 1)})

    def test_split_group_diagram(self):
        diagram = self.build('split_group.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (3, 0), 'D': (2, 0),
                                    'E': (4, 0)})

    def test_peer_network_diagram(self):
        diagram = self.build('peer_network.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (1, 2)})

    def test_peer_network2_diagram(self):
        diagram = self.build('peer_network2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1)})

    def test_peer_network3_diagram(self):
        diagram = self.build('peer_network3.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2)})

    def test_peer_network4_diagram(self):
        diagram = self.build('peer_network4.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 1),
                                    'C': (1, 2)})

    def test_peer_network5_diagram(self):
        diagram = self.build('peer_network5.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1)})

    def test_peer_network_branched_diagram(self):
        diagram = self.build('peer_network_branched.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (1, 2), 'D': (1, 3)})

    def test_same_peer_network_diagram(self):
        diagram = self.build('same_peer_network.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (1, 2), 'D': (0, 3)})

    def test_group_and_peer_network_diagram(self):
        diagram = self.build('group_and_peer_network.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (1, 2)})

    def test_peer_network_in_same_node_diagram(self):
        with self.assertRaises(RuntimeError):
            self.build('errors/peer_network_in_same_node.diag')



# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages

sys.path.insert(0, 'src')
import nwdiag

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Software Development :: Documentation",
    "Topic :: Text Processing :: Markup",
]
test_requires = ['nose',
                 'pep8>=1.3',
                 'reportlab',
                 'docutils']

# only for Python2.6
if sys.version_info > (2, 6) and sys.version_info < (2, 7):
    test_requires.append('unittest2')

setup(
    name='nwdiag',
    version=nwdiag.__version__,
    description='nwdiag generates network-diagram image from text',
    long_description=open("README.rst").read(),
    classifiers=classifiers,
    keywords=['diagram', 'generator'],
    author='Takeshi Komiya',
    author_email='i.tkomiya at gmail.com',
    url='http://blockdiag.com/',
    download_url='http://pypi.python.org/pypi/nwdiag',
    license='Apache License 2.0',
    py_modules=[
        'nwdiag_sphinxhelper',
        'rackdiag_sphinxhelper',
        'packetdiag_sphinxhelper',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['buildout.cfg']},
    include_package_data=True,
    install_requires=[
        'setuptools',
        'blockdiag>=1.5.0',
        # -*- Extra requirements: -*-
    ],
    extras_require=dict(
        testing=test_requires,
        pdf=[
            'reportlab',
        ],
        rst=[
            'docutils',
        ],
    ),
    test_suite='nose.collector',
    tests_require=test_requires,
    entry_points="""
       [console_scripts]
       nwdiag = nwdiag.command:main
       rackdiag = rackdiag.command:main
       packetdiag = packetdiag.command:main

       [blockdiag_noderenderer]
       _packet_node = packetdiag.noderenderers
    """,
)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import NodeShape
from blockdiag.noderenderer import install_renderer


class PacketNode(NodeShape):
    def render_shape(self, drawer, _, **kwargs):
        # draw outline
        box = self.metrics.cell(self.node).box
        if kwargs.get('shadow'):
            pass
        else:
            drawer.rectangle(box, fill=self.node.color,
                             outline=self.node.color)
            if self.node.background:
                drawer.loadImage(self.node.background, self.textbox)

            self.line(drawer, box.topleft, box.topright, False)
            self.line(drawer, box.topright, box.bottomright,
                      self.node.separated_right)
            self.line(drawer, box.bottomright, box.bottomleft, False)
            self.line(drawer, box.bottomleft, box.topleft,
                      self.node.separated_left)

    def line(self, drawer, _from, _to, is_dashed):
        if is_dashed:
            style = 'dashed'
        else:
            style = self.node.style

        drawer.line((_from, _to), fill=self.node.linecolor, style=style)


def setup(self):
    install_renderer('_packet_node', PacketNode)



# -*- coding: utf-8 -*-

import os
from nwdiag.builder import ScreenNodeBuilder
from nwdiag.parser import parse_file
from blockdiag.tests.utils import BuilderTestCase as BlockdiagBuilderTestCase


class BuilderTestCase(BlockdiagBuilderTestCase):
    def build(self, filename):
        basedir = os.path.dirname(__file__)
        pathname = os.path.join(basedir, 'diagrams', filename)
        return ScreenNodeBuilder.build(parse_file(pathname))



# -*- coding: utf-8 -*-

from __future__ import print_function

import sys
if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

import os
from rackdiag.builder import ScreenNodeBuilder
from rackdiag.parser import parse_string


def build(filename):
    def decorator(fn):
        def func(self):
            testdir = os.path.dirname(__file__)
            pathname = os.path.join(testdir, 'diagrams', filename)

            code = open(pathname).read()
            tree = parse_string(code)
            self.diagram = ScreenNodeBuilder.build(tree)

            return fn(self)

        func.__name__ = fn.__name__
        return func

    return decorator


class TestBuildDiagram(unittest.TestCase):
    def assertAttributes(self, **attributes):
        for rack in self.diagram.racks:
            for node in rack.nodes:
                print(node, node.label)

                for key, attribute in attributes.items():
                    value = attribute.get(node.label)
                    self.assertEqual(value, getattr(node, key))

    @build('simple.diag')
    def test_simple(self):
        positions = {'server 1': (0, 41),
                     'server 2': (0, 40),
                     'server 3': (0, 39),
                     'server 4': (0, 38)}

        self.assertAttributes(xy=positions)

    @build('autonumber.diag')
    def test_autonumber(self):
        positions = {'server 1': (0, 41),
                     'server 2': (0, 40),
                     'server 3': (0, 38),
                     'server 4': (0, 37)}

        self.assertAttributes(xy=positions)

    @build('multi_rackitem.diag')
    def test_multi_rackitem(self):
        positions = {'UPS 1': (0, 9), 'UPS 2': (6, 9),
                     'server 1': (0, 8), 'server 2': (3, 7),
                     'server 3': (0, 6), 'server 4': (3, 5),
                     'server 5': (6, 5), 'server 6': (9, 5),
                     'server 7': (0, 2), 'server 8': (4, 3),
                     'server 9': (8, 3)}
        colheights = {'UPS 1': 3, 'UPS 2': 3,
                      'server 1': 1, 'server 2': 2,
                      'server 3': 2, 'server 4': 2,
                      'server 5': 2, 'server 6': 2,
                      'server 7': 3, 'server 8': 1,
                      'server 9': 1}

        self.assertAttributes(xy=positions, colheight=colheights)

    @build('multi_racks.diag')
    def test_multi_racks(self):
        positions = {'server 1': (0, 41), 'server 2': (0, 40),
                     'server 3': (0, 39), 'server 4': (0, 38),
                     'server 5': (1, 41), 'server 6': (1, 40),
                     'server 7': (1, 39), 'server 8': (1, 38)}

        self.assertAttributes(xy=positions)
